/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* BOOLEAN:OBJECT (../libadwaita/src/adw-marshalers.list:28) */
extern
void adw_marshal_BOOLEAN__OBJECT (GClosure     *closure,
                                  GValue       *return_value,
                                  guint         n_param_values,
                                  const GValue *param_values,
                                  gpointer      invocation_hint,
                                  gpointer      marshal_data);
extern
void adw_marshal_BOOLEAN__OBJECTv (GClosure *closure,
                                   GValue   *return_value,
                                   gpointer  instance,
                                   va_list   args,
                                   gpointer  marshal_data,
                                   int       n_params,
                                   GType    *param_types);

/* BOOLEAN:STRING (../libadwaita/src/adw-marshalers.list:29) */
extern
void adw_marshal_BOOLEAN__STRING (GClosure     *closure,
                                  GValue       *return_value,
                                  guint         n_param_values,
                                  const GValue *param_values,
                                  gpointer      invocation_hint,
                                  gpointer      marshal_data);
extern
void adw_marshal_BOOLEAN__STRINGv (GClosure *closure,
                                   GValue   *return_value,
                                   gpointer  instance,
                                   va_list   args,
                                   gpointer  marshal_data,
                                   int       n_params,
                                   GType    *param_types);

/* BOOLEAN:VOID (../libadwaita/src/adw-marshalers.list:30) */
extern
void adw_marshal_BOOLEAN__VOID (GClosure     *closure,
                                GValue       *return_value,
                                guint         n_param_values,
                                const GValue *param_values,
                                gpointer      invocation_hint,
                                gpointer      marshal_data);
extern
void adw_marshal_BOOLEAN__VOIDv (GClosure *closure,
                                 GValue   *return_value,
                                 gpointer  instance,
                                 va_list   args,
                                 gpointer  marshal_data,
                                 int       n_params,
                                 GType    *param_types);

/* INT:POINTER (../libadwaita/src/adw-marshalers.list:31) */
extern
void adw_marshal_INT__POINTER (GClosure     *closure,
                               GValue       *return_value,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint,
                               gpointer      marshal_data);
extern
void adw_marshal_INT__POINTERv (GClosure *closure,
                                GValue   *return_value,
                                gpointer  instance,
                                va_list   args,
                                gpointer  marshal_data,
                                int       n_params,
                                GType    *param_types);

/* OBJECT:VOID (../libadwaita/src/adw-marshalers.list:32) */
extern
void adw_marshal_OBJECT__VOID (GClosure     *closure,
                               GValue       *return_value,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint,
                               gpointer      marshal_data);
extern
void adw_marshal_OBJECT__VOIDv (GClosure *closure,
                                GValue   *return_value,
                                gpointer  instance,
                                va_list   args,
                                gpointer  marshal_data,
                                int       n_params,
                                GType    *param_types);

/* VOID:BOOLEAN (../libadwaita/src/adw-marshalers.list:33) */
#define adw_marshal_VOID__BOOLEAN	g_cclosure_marshal_VOID__BOOLEAN
#define adw_marshal_VOID__BOOLEANv	g_cclosure_marshal_VOID__BOOLEANv

/* VOID:DOUBLE (../libadwaita/src/adw-marshalers.list:34) */
#define adw_marshal_VOID__DOUBLE	g_cclosure_marshal_VOID__DOUBLE
#define adw_marshal_VOID__DOUBLEv	g_cclosure_marshal_VOID__DOUBLEv

/* VOID:DOUBLE,DOUBLE (../libadwaita/src/adw-marshalers.list:35) */
extern
void adw_marshal_VOID__DOUBLE_DOUBLE (GClosure     *closure,
                                      GValue       *return_value,
                                      guint         n_param_values,
                                      const GValue *param_values,
                                      gpointer      invocation_hint,
                                      gpointer      marshal_data);
extern
void adw_marshal_VOID__DOUBLE_DOUBLEv (GClosure *closure,
                                       GValue   *return_value,
                                       gpointer  instance,
                                       va_list   args,
                                       gpointer  marshal_data,
                                       int       n_params,
                                       GType    *param_types);

/* VOID:DOUBLE,UINT (../libadwaita/src/adw-marshalers.list:36) */
extern
void adw_marshal_VOID__DOUBLE_UINT (GClosure     *closure,
                                    GValue       *return_value,
                                    guint         n_param_values,
                                    const GValue *param_values,
                                    gpointer      invocation_hint,
                                    gpointer      marshal_data);
extern
void adw_marshal_VOID__DOUBLE_UINTv (GClosure *closure,
                                     GValue   *return_value,
                                     gpointer  instance,
                                     va_list   args,
                                     gpointer  marshal_data,
                                     int       n_params,
                                     GType    *param_types);

/* VOID:ENUM (../libadwaita/src/adw-marshalers.list:37) */
#define adw_marshal_VOID__ENUM	g_cclosure_marshal_VOID__ENUM
#define adw_marshal_VOID__ENUMv	g_cclosure_marshal_VOID__ENUMv

/* VOID:INT (../libadwaita/src/adw-marshalers.list:38) */
#define adw_marshal_VOID__INT	g_cclosure_marshal_VOID__INT
#define adw_marshal_VOID__INTv	g_cclosure_marshal_VOID__INTv

/* VOID:OBJECT (../libadwaita/src/adw-marshalers.list:39) */
#define adw_marshal_VOID__OBJECT	g_cclosure_marshal_VOID__OBJECT
#define adw_marshal_VOID__OBJECTv	g_cclosure_marshal_VOID__OBJECTv

/* VOID:OBJECT,INT (../libadwaita/src/adw-marshalers.list:40) */
extern
void adw_marshal_VOID__OBJECT_INT (GClosure     *closure,
                                   GValue       *return_value,
                                   guint         n_param_values,
                                   const GValue *param_values,
                                   gpointer      invocation_hint,
                                   gpointer      marshal_data);
extern
void adw_marshal_VOID__OBJECT_INTv (GClosure *closure,
                                    GValue   *return_value,
                                    gpointer  instance,
                                    va_list   args,
                                    gpointer  marshal_data,
                                    int       n_params,
                                    GType    *param_types);

/* VOID:STRING (../libadwaita/src/adw-marshalers.list:41) */
#define adw_marshal_VOID__STRING	g_cclosure_marshal_VOID__STRING
#define adw_marshal_VOID__STRINGv	g_cclosure_marshal_VOID__STRINGv

/* VOID:UINT (../libadwaita/src/adw-marshalers.list:42) */
#define adw_marshal_VOID__UINT	g_cclosure_marshal_VOID__UINT
#define adw_marshal_VOID__UINTv	g_cclosure_marshal_VOID__UINTv

/* VOID:VOID (../libadwaita/src/adw-marshalers.list:43) */
#define adw_marshal_VOID__VOID	g_cclosure_marshal_VOID__VOID
#define adw_marshal_VOID__VOIDv	g_cclosure_marshal_VOID__VOIDv


G_END_DECLS
