#pragma once

#include <adwaita.h>

G_BEGIN_DECLS

#define ADW_TYPE_DEMO_PAGE_BOTTOM_SHEETS (adw_demo_page_bottom_sheets_get_type())

G_DECLARE_FINAL_TYPE (AdwDemoPageBottomSheets, adw_demo_page_bottom_sheets, ADW, DEMO_PAGE_BOTTOM_SHEETS, AdwBin)

G_END_DECLS
