#pragma once

#include <adwaita.h>

G_BEGIN_DECLS

#define ADW_TYPE_DEMO_PAGE_STYLES (adw_demo_page_styles_get_type())

G_DECLARE_FINAL_TYPE (AdwDemoPageStyles, adw_demo_page_styles, ADW, DEMO_PAGE_STYLES, AdwBin)

G_END_DECLS
